/*
 * Copyright(c) 1997, 1998 by Jun-ichiro Itoh. All rights reserved.
 * Freely redistributable unless otherwise noted.  Absolutely no warranty.
 *
 * author contact: Jun-ichiro itojun Itoh <itojun@mt.cs.keio.ac.jp>
 * $Id: main.c,v 1.6 1998/09/22 07:38:49 itojun Exp $
 */

#include <common.h>

#if HAVE_TERMIOS_H
# include <termios.h>
#else
/*# if HAVE_TERMIO_H*/
# if 0
#  include <termio.h>
# else
#  if HAVE_SYS_IOCTL_H
#   include <sys/ioctl.h>
#  endif
#  include <sgtty.h>
# endif
#endif
#include <setjmp.h>
#include <err.h>
#include <sys/stat.h>

/* serial lines */
static int baudrateflag = 115200;
static char *n0;
static int fd0;

static int dumpflag = 0;
static int dumpmaxlen = 20;
static int debugflag = 0;
static int openflag = O_WRONLY | O_TRUNC | O_CREAT;
static int verbose = 0;
static char *pathname = NULL;

#define	SMALLWAIT	200*1000	/*usec*/

#define dprintf(x)						\
	{ if (debugflag) {					\
		fprintf(stderr, __FILE__ ":%d: ", __LINE__);	\
		fprintf x;					\
	  }							\
	}

/* uid */
static uid_t uid, euid;
static gid_t gid, egid;
static int uidswapped = 0;

/* uucp locking */
static int useuucplock = 1;
extern int uu_lock __P((char *));
extern int uu_unlock __P((char *));

/* error handling */
jmp_buf fatal;

void
daemonuid()
{
	if (uidswapped == 0)
		return;

#ifdef HAVE_SETREUID
	setreuid(uid, euid);
	setregid(gid, egid);
#else
	setuid(uid);
	seteuid(euid);
	setgid(gid);
	setegid(egid);
#endif
	uidswapped = 0;
}

void
useruid()
{
	if (uidswapped == 1)
		return;

#ifdef HAVE_SETREUID
	setregid(egid, gid);
	setreuid(euid, uid);
#else
	setgid(egid);
	setegid(gid);
	setuid(euid);
	seteuid(uid);
#endif
	uidswapped = 1;
}

static int
uucplock(name)
	char *name;
{
	int ret;

	if (useuucplock) {
		daemonuid();
		ret = uu_lock(rindex(name, '/') + 1);
		useruid();
	} else
		ret = 0;
	return ret;
}

static int
uucpunlock(name)
	char *name;
{
	int ret;

	if (useuucplock) {
		daemonuid();
		ret = uu_unlock(rindex(name, '/') + 1);
		useruid();
	} else
		ret = 0;
	return ret;
}

static char *
canondev(name)
	char *name;
{
	int fd;
	char *p;

	daemonuid();

	p = malloc(strlen(name) + 1);
	if (!p) {
		perror("malloc");
		goto back;
	}
	strcpy(p, name);
	fd = open(p, O_RDWR|O_NONBLOCK);
	if (0 <= fd) {
		close(fd);
		goto back;
	}

	p = realloc(p, strlen(name) + 6);
	if (!p) {
		perror("realloc");
		goto back;
	}
	strcpy(p, "/dev/");
	strcat(p, name);
	fd = open(p, O_RDWR|O_NONBLOCK);
	if (0 <= fd) {
		close(fd);
		goto back;
	}

	free(p);
	p = NULL;
back:
	useruid();
	return p;
}

speed_t
baudconv(int baud)
{
#define	BAUDCASE(x)	case (x): { ret = B##x; break; }
	speed_t ret;

	ret = (speed_t) baud;
	switch (baud) {
	/* POSIX defined baudrates */
	BAUDCASE(0);	/*is it meaningful? */
	BAUDCASE(50);
	BAUDCASE(75);
	BAUDCASE(110);
	BAUDCASE(134);
	BAUDCASE(150);
	BAUDCASE(200);
	BAUDCASE(300);
	BAUDCASE(600);
	BAUDCASE(1200);
	BAUDCASE(1800);
	BAUDCASE(2400);
	BAUDCASE(4800);
	BAUDCASE(9600);
	BAUDCASE(19200);
	BAUDCASE(38400);

	/* non POSIX values */
#ifdef B7200
	BAUDCASE(7200);
#endif
#ifdef B14400
	BAUDCASE(14400);
#endif
#ifdef B28800
	BAUDCASE(28800);
#endif
#ifdef B57600
	BAUDCASE(57600);
#endif
#ifdef B115200
	BAUDCASE(115200);
#endif
#ifdef B230400
	BAUDCASE(230400);
#endif

	/* last resort */
	default:
		fprintf(stderr, "no defined value for baudrate %d found, "
			"use the value without conversion\n", baud);
	}

	return ret;
#undef BAUDCASE
}

int
setbaud(fd, baud)
	int fd;
	int baud;
{
#if HAVE_TERMIOS_H
	/* termios */
	struct termios tio;

	if (tcgetattr(fd, &tio) < 0) {
		perror("tcgetattr");
		return 1;
	}
	tio.c_iflag = 0;
	tio.c_oflag = 0;
	tio.c_cflag = CS8 | CREAD | CLOCAL;
	tio.c_lflag = 0;
	tio.c_cc[VMIN] = 1;
	tio.c_cc[VTIME] = 5;
	cfsetispeed(&tio, baudconv(baud));
	cfsetospeed(&tio, baudconv(baud));
	if (tcsetattr(fd, TCSANOW, &tio) < 0) {
		perror("tcsetattr");
		return 1;
	}
#else
/*#if HAVE_TERMIO_H*/
# if 0
	/* termio */
# else
	/* sgtty */
	struct sgttyb ttyb;

	if (ioctl(fd, TIOCGETP, &ttyb) < 0) {
		perror("ioctl(TIOCGETP)");
		return 1;
	}
	ttyb.sg_ispeed = baud;
	ttyb.sg_ospeed = baud;
	ttyb.sg_flags = 0;
	if (ioctl(fd, TIOCSETP, &ttyb) < 0) {
		perror("ioctl(TIOCSETP)");
		return 1;
	}
# endif
#endif

	dprintf((stderr, "baudrate set to %d\n", baud));
	return 0;
}

void
dump_stream(dir, buf, len)
	int dir;
	u_char *buf;
	int len;
{
	size_t i;
	int truncate;

	if (!dumpflag)
		return;

	truncate = 0;
	if (dumpmaxlen < len) {
		len = dumpmaxlen;
		truncate = 1;
	}

	dir &= 0xff;
	if (dir == '>')
		fprintf(stderr, "crosspad>cpu: ");
	else
		fprintf(stderr, "cpu>crosspad: ");

	for (i = 0; i < len; i++)
		fprintf(stderr, "%02x ", buf[i] & 0xff);
	if (truncate)
		fprintf(stderr, "...");
	fprintf(stderr, "\n");
}

/*------------------------------------------------------------*/

u_char crosspad_buf[1024 * 4];
size_t crosspad_len;

void
crosspad_init()
{
	/* initialize buffers */
	crosspad_len = 0;
}

int
crosspad_wait(rlen)
	int rlen;
{
	fd_set rdfd;
	int maxfd;
	int len;
	struct timeval t;

	while (crosspad_len < rlen) {
		/* obtain chars from the crosspad */
		FD_ZERO(&rdfd);
		FD_SET(fd0, &rdfd);
		maxfd = fd0;
		/* timeout: 5sec */
		t.tv_sec = 5;
		t.tv_usec = 0;
		switch (select(maxfd + 1, &rdfd, NULL, NULL, &t)) {
		case -1:
			if (errno == EINTR)
				break;
			perror("select");
			exit(1);
		case 0:
			dprintf((stderr, "read timeout.\n"));
			return 1;
		default:
			break;
		}

		if (FD_ISSET(fd0, &rdfd)) {
			len = read(fd0, crosspad_buf + crosspad_len,
				sizeof(crosspad_buf) - crosspad_len);
			crosspad_len += len;
		} else {
			dprintf((stderr, "something wrong in crosspad_get\n"));
			return 1;
		}
	}
	dump_stream('>', crosspad_buf, crosspad_len);

	return 0;
}

int
crosspad_get(buf, len)
	u_char *buf;
	int len;
{
	switch (crosspad_wait(len)) {
	case 1:
		dprintf((stderr, "timeouted in crosspad_get\n"));
		fprintf(stderr, "crosspad not ready.\n");
		longjmp(fatal, 1);
	case 0:
		break;
	}

	if (crosspad_len < len)
		abort();

	memcpy(buf, crosspad_buf, len);
	if (crosspad_len - len)
		memmove(crosspad_buf, crosspad_buf + len, crosspad_len - len);
	crosspad_len -= len;

	return 0;
}

int
crosspad_put(buf, len)
	u_char *buf;
	int len;
{
	int wlen;
	int i;

	/* Thanks goes to Thierry Bousch */
	for (i = 0; i < len; i++) {
		wlen = write(fd0, &buf[i], 1);
		if (wlen != 1) {
			dprintf((stderr, "failed in crosspad_put\n"));
			return 1;
		}
		usleep(20*1000);
#if HAVE_TERMIOS_H
		/*termios*/
		tcdrain(fd0);
#else
/*#if HAVE_TERMIO_H*/
#if 0
		/*termio - yet to be done*/
#else
		/*sgtty*/
		ioctl(fd0, TCDRAIN, 0);
#endif
#endif
	}

	dump_stream('<', buf, len);
	return 0;
}

/*------------------------------------------------------------*/

int
crosspad_sync()
{
	u_char buf[10];
	time_t timeout;
	u_int sum;
	int i;

	while (1) {
		crosspad_get(buf, 1);
		if (buf[0] != 0x02) {
			dprintf((stderr, "ignore non sync byte "
				"in crosspad_sync\n"));
			continue;
		}

		crosspad_get(buf + 1, 5);
		sum = 0x00;
		for (i = 1; i < 5; i++)
			sum ^= buf[i];
		if ((sum & 0xff)!= buf[5]) {
			dprintf((stderr, "sync packet checksum error(%02x %02x",
				sum & 0xff, buf[5]));
		}
		break;
	}

	return ntohl(*(u_int *)&buf[1]);
}

int
crosspad_sendack()
{
	crosspad_put("\006", 1);
	return 0;
}

int
crosspad_skip(n)
	int n;
{
	u_char buf[10];

	crosspad_get(buf, n);
	return 0;
}

int
crosspad_download(fd, len)
	int fd;
	int len;
{
#define DOWNLOADLEN	100
	u_char buf[DOWNLOADLEN];
	u_char sector;
	int slen;	/* sector length */
	int l, l2, l3;

	sector = 0;
	l = len;
	while (0 < l) {
		crosspad_get(buf, 5);
		if (buf[0] != 0x01 || buf[1] != sector
		 || buf[1] + buf[2] != 0xff) {
			fprintf(stderr, "out of sync\n");
			exit(1);
		}
		slen = (u_short)buf[3] << 8 | buf[4];
		l2 = slen;
		while (0 < l2) {
			l3 = l2 < DOWNLOADLEN ? l2 : DOWNLOADLEN;
			crosspad_get(buf, l3);
			write(fd, buf, l3);
			l2 -= l3;

			if (verbose && !dumpflag) {
				fprintf(stderr,
					"download sector %3d: %d/%d  \r",
					sector, (len - l) + (slen - l2), len);
			}
		}
		l -= slen;

		crosspad_get(buf, 1);	/*junk byte - checksum?*/

		if (verbose) {
			fprintf(stderr, "download sector %3d: %d/%d  \n",
				sector, len - l, len);
		}

		crosspad_sendack();
		sector++;
	}
	if (verbose)
		fprintf(stderr, "download finished\n");

	return 0;
}

void
mainloop()
{
	int fd;
	int len;

	switch (setjmp(fatal)) {
	case 0:
		break;
	default:
		goto fatalerr;
	}

	fd = open(pathname, openflag, 0644);
	if (fd < 0) {
		warn("open");
		goto fatalerr;
	}

	/*initial value*/
	if (setbaud(fd0, baudrateflag)) {
		fprintf(stderr, "can't set baudrate\n");
		goto fatalerr;
	}

	crosspad_init();

	usleep(SMALLWAIT);

	if (verbose)
		fprintf(stderr, "waiting for input...\n");
	len = crosspad_sync();
	fprintf(stderr, "downloading %d bytes\n", len);
	crosspad_sendack();
	crosspad_download(fd, len);
	close(fd);
	crosspad_skip(1);

	return;

fatalerr:
	/* cleanup() will be called automatically */
	exit(1);
}

void
usage()
{
	fprintf(stderr, "usage: padload [-dDiuv] [-b baudrate] [-o path] port\n");
}

void
cleanup()
{
	close(fd0);
	uucpunlock(n0);
}

void
main(argc, argv)
	int argc;
	char **argv;
{
	int opt;
	extern char *optarg;
	char pathbuf[MAXPATHLEN];

	uid = getuid();
	euid = geteuid();
	gid = getgid();
	egid = getegid();
	useruid();

	useuucplock = 1;

	while ((opt = getopt(argc, argv, "b:dDio:uv")) != -1) {
		switch (opt) {

		/* controls */
		case 'b':
			baudrateflag = atoi(optarg);
			break;
		case 'd':
			dumpflag = 1;
			break;
		case 'D':
			debugflag = 1;
			break;
		case 'i':
			openflag = O_WRONLY | O_EXCL | O_CREAT;
			break;
		case 'o':
			pathname = optarg;
			break;
		case 'u':
			useuucplock = 0;
			break;
		case 'v':
			verbose = 1;
			break;
		default:
			usage();
			exit(1);
		}
	}
	argc -= optind;
	argv += optind;

	if (1 < argc) {
		fprintf(stderr, "invalid argument\n");
		usage();
		exit(1);
	} else if (argc == 0) {
		fprintf(stderr, "no device specified\n");
		usage();
		exit(1);
	}

	if (!pathname) {
		time_t t;
		time(&t);
		snprintf(pathbuf, sizeof(pathbuf), "/tmp/%ld.cpd", t);
		pathname = pathbuf;

		fprintf(stderr, "generating output to %s\n", pathname);
	}

	n0 = canondev(argv[0]);
	if (!n0) {
		fprintf(stderr, "inappropriate device specified, "
			"or device permission error\n");
		exit(1);
	}

	if (uucplock(n0) < 0) {
		fprintf(stderr, "device %s in use\n", n0);
		exit(1);
	}

	daemonuid();
	fd0 = open(n0, O_RDWR|O_NDELAY);
	if (fd0 < 0) {
		perror("open");
		exit(1);
	}
	useruid();

	atexit(cleanup);

	mainloop();

	exit(0);
}
